<?php  
// fschuett@gymnasium-himmelsthuer.de, 2012

require_once("$CFG->dirroot/enrol/enrol.class.php");

class enrolment_plugin_lml {

    var $log;    
    
    //idnumber: Moodle "Kurs-ID"
    var $enrol_localcoursefield = 'idnumber';
    
    var $teacher_array=Array();

    var $ldap_userfilter = '';
/**
 * constructor since php5
*/
function __construct() {
	$object=get_record('config_plugins','plugin','auth/ldap','name','objectclass');
	if(empty($object) || empty($object->value)) {
		$this->ldap_userfilter='(objectClass=*)';
	} else if (stripos($object->value, 'objectClass=') === 0) {
		$this->ldap_userfilter = '('.$object->value.')';
	} else if (stripos($object->value, '(') !== 0) {
		$this->ldap_userfilter = '(objectClass='.$object->value.')';
	} else {
		$this->ldap_userfilter = $object->value;
	}
	//error_log('[ENROL_LML __construct] ldap_userfilter['.$this->ldap_userfilter.']');
}

/**
 * legacy constructor for php4
*/
function enrolment_plugin_lml() {
	return $this->__construct();
}

/**
 * This function syncs a user's teacher group enrolments with those on the LML server.
 */
function setup_teacher_enrolments($ldap_connection,&$user,$isteacher) {
    global $CFG;

    //error_log('[ENROL_LML] setup_teacher_enrolments called');

    //Get the role in question
    $role = get_record('role','shortname',$CFG->{'enrol_lml_teachers_role'});

    
    //Get the LDAP teacher role
    $this->enrol_lml_bind($ldap_connection);
    
    //Correct the Moodle enrolments
    // find the courses
    $courses_objs = $this->moodle_courses_for_group($CFG->{'enrol_lml_teachers_course_id'});
    //error_log('[ENROL_LML] setup_teacher_enrolments teacher course objects '.print_r($courses_objs,true).')');
    if(!empty($courses_objs)) { //there exist courses
	foreach($courses_objs as $course_obj){
	    // deal with enrolment in the moodle db
	    $context = get_context_instance(CONTEXT_COURSE, $course_obj->id);
	    if($isteacher==true){
		if (!get_record('role_assignments', 'roleid', $role->id, 'userid', $user->id, 'contextid', $context->id)) {
		    error_log("[ENROL_LML setup_teacher_enrolments] Assigning role '{$role->name}' to {$user->id} ({$user->username}) in course {$course_obj->id} ({$course_obj->shortname})");
		    if (!role_assign($role->id, $user->id, 0, $context->id, 0, 0, 0, 'lml')){
			error_log("[ENROL_LML setup_teacher_enrolments] Failed to assign role '{$role->name}' to $user->id ($user->username) into course $course_obj->id ($course_obj->shortname)");
		    } 
		} else {
		    //error_log("[ENROL_LML setup_teacher_enrolments] Role '{$role->name}' already assigned to {$user->id} ({$user->username}) in course {$course_obj->id} ({$course_obj->shortname})");
		}
	    }
	    else {
		if (get_record('role_assignments', 'roleid', $role->id, 'userid', $user->id, 'contextid', $context->id)) {
		    //error_log("[ENROL_LML setup_teacher_enrolments] Unassigning role '{$role->name}' to {$user->id} ({$user->username}) in course {$course_obj->id} ({$course_obj->shortname})");
		    if (!role_unassign($role->id, $user->id, 0, $context->id, 0, 0, 0, 'lml')){
			error_log("[ENROL_LML setup_teacher_enrolments] Failed to unassign role '{$role->name}' to $user->id ($user->username) into course $course_obj->id ($course_obj->shortname)");
		    } 
		} else {
		    //error_log("[ENROL_LML setup_teacher_enrolments] Role '{$role->name}' already unassigned to {$user->id} ({$user->username}) in course {$course_obj->id} ({$course_obj->shortname})");
		}
	    }
	}    
    }
    

    return true;

}

/**
 * This function syncs a student's group enrolments with the classes and projects on the LML server.
 */
function setup_student_enrolments($ldap_connection,&$user) {
    global $CFG;

    //error_log('[ENROL_LML] setup_student_enrolments called');

    //Get the role in question
    $role = get_record('role','shortname',$CFG->{'enrol_lml_student_role'});

    
    //Get the LDAP teacher role
    $this->enrol_lml_bind($ldap_connection);
    
    $isteacher = $this->is_teacher($ldap_connection,$user->idnumber);
    if($isteacher==true){
      //error_log($user->idnumber.' is a teacher. Nothing to do.');
      return true;
    }
    
    // Get the entire list of role assignments for role that currently exist for this user.
    $roleassignments = get_records('role_assignments', 'userid', $user->id);
    if (!$roleassignments) {
        $roleassignments = array();
    }
    
    // Get the ones that came from LDAP
    $lml_assignments = array_filter($roleassignments, create_function('$x', 'return $x->enrol == \'lml\';'));

    //Find the external student enrolments
    $ldapcourses = $this->find_ext_student_enrolments($ldap_connection,$user->idnumber,$role);
    //correct array
    //error_log('[ENROL_LML setup_student_enrolments] '.count($ldapcourses).' course(s).');
    
    $ldapcourses = addslashes_recursive($ldapcourses);
    
    //find all courses with external enrolments
    foreach($ldapcourses as $ldapcourse) {
        $ldapcourse = trim($ldapcourse);
        if(empty($ldapcourse)) {
	    continue;
	}
        //error_log('[ENROL_LML setup_student_enrolments] course '.$ldapcourse); 

	//Correct the Moodle enrolments
	// find the courses
	$courses_objs = $this->moodle_courses_for_group($ldapcourse);
	if(empty($courses_objs)) {
	    error_log("[ENROL_LML setup_student_enrolments] user $user->username enrolled to a nonexistant course $ldapcourse \n");
	    continue;
	}

	//enrol in moodle into this courses
	foreach($courses_objs as $course_obj){
	    // deal with enrolment in the moodle db
	    $context = get_context_instance(CONTEXT_COURSE, $course_obj->id);
	    //$courseroles = get_user_roles($context, $user->id);
	
	    if (!get_record('role_assignments', 'roleid', $role->id, 'userid', $user->id, 'contextid', $context->id)) {
		//error_log("[ENROL_LML setup_student_enrolments] Assigning role '{$role->name}' to {$user->id} ({$user->username}) in course {$course_obj->id} ({$course_obj->shortname})");
		if (!role_assign($role->id, $user->id, 0, $context->id, 0, 0, 0, 'lml')){
		    error_log("[ENROL_LML setup_student_enrolments] Failed to assign role '{$role->name}' to $user->id ($user->username) into course $course_obj->id ($course_obj->shortname)");
		} 
	    } else {
		//error_log("[ENROL_LML setup_student_enrolments] Role '{$role->name}' already assigned to {$user->id} ({$user->username}) in course {$course_obj->id} ({$course_obj->shortname})");
	    }

	    // Remove the role from the list we created earlier.  This
	    // way we can find those roles that are no longer required.
	    foreach($lml_assignments as $key => $roleassignment) {
		if ($roleassignment->roleid == $role->id
		  && $roleassignment->contextid == $context->id) {
		    unset($lml_assignments[$key]);
		    break;
		}
	    }
	}
    }
    
    // ok, if there's any thing still left in the $roleassignments array we
    // made at the start, we want to remove any of the ldap ones.
    foreach ($lml_assignments as $ra) {
        if($ra->enrol === 'lml') {
            //error_log("Unassigning role_assignment with id '{$ra->id}' from user {$user->id} ({$user->username})");
            if(!role_unassign($ra->roleid, $user->id, 0, $ra->contextid, 'lml')){
		error_log("[ENROL_LML setup_student_enrolments] Failed to unassign role_assignment with id '{$ra->id}' from user {$user->id} ({$user->username})");
	    }
        }
    }
    return true;

}

/**
 * This function syncs a user's enrolments with those on the LML server.
 */
function setup_enrolments(&$user) {
    global $CFG;

    //error_log('[ENROL_LML] setup_enrolments called');

    // Connect to the external database
    $ldap_connection = $this->enrol_lml_connect();
    if (!$ldap_connection) {
        @ldap_close($ldap_connection);
        notify("[ENROL_LML] LML-module cannot connect to server: {$CFG->enrol_lml_host_url}");
        return false;
    }
    
    // we are connected OK, continue...
    
    // Make sure the config settings have been upgraded.
    $this->check_legacy_config();

    $isteacher = $this->is_teacher($ldap_connection,$user->idnumber);
    
    //Correct the teacher enrollments
    if(!$this->setup_teacher_enrolments($ldap_connection,$user,$isteacher)){
      notify("[ENROL_LML setup_enrolments] setup teacher enrollments failed for user ".$user->idnumber);
      return false;
    }
    
    //Correct the student enrollments
    if(!$this->setup_student_enrolments($ldap_connection,$user)){
      notify("[ENROL_LML setup_enrolments] setup student enrollments failed for user ".$user->idnumber);
      return false;
    }
    
    //autocreate/autoremove teacher category
    if($CFG->enrol_lml_teachers_category_autocreate || $CFG->enrol_lml_teachers_category_autoremove){
	if(!$this->setup_teacher_category($user,$isteacher)){
	    notify("[ENROL_LML setup_teacher_category] failed for user ".$user->idnumber);
	    return false;
	}
    }
    
    @ldap_close($ldap_connection);

    //error_log('[ENROL_LML] finished with setup_enrolments');
    
    return true;
}

/// sync teachers enrolments with lml 
function sync_teacher_enrolments($ldap_connection, $role, $enrol) {
    global $CFG;

    //error_log('[ENROL_LML sync_teacher_enrolments] start');
    if(empty($this->teacher_array)) {
	$this->init_teacher_array($ldap_connection);
    }
    
    //only the teachers role will be used
    if($role->shortname != $CFG->enrol_lml_teachers_role) {
	return true;
    }
    
    //find course objs
    $idnumber = $CFG->{'enrol_lml_teachers_course_id'};
    print "==Syncing $idnumber\n";
    
    // do courses exist in moodle? 
    $course_objs = false;
	$course_objs = $this->moodle_courses_for_group( $idnumber );
    
      foreach($course_objs as $course_obj){
	// enrol&unenrol if required
	if($enrol && is_object($course_obj)){
	  // Get a context object.
	  $context = get_context_instance(CONTEXT_COURSE, $course_obj->id);
	  // pull the ldap membership into a nice array
	  // this is an odd array -- mix of hash and array -- 
	  $ldapmembers=$this->teacher_array;
          //error_log('[ENROL_LML sync_teacher_enrolments] teachers count '.$ldapmembers['count']);

	  unset($ldapmembers['count']);
	  $ldapmembers = addslashes_recursive($ldapmembers);
	  
	  // prune old lml enrolments
	  // hopefully they'll fit in the max buffer size for the RDBMS
	  $sql = '
	      SELECT enr.userid AS user, 1
	      FROM '.$CFG->prefix.'role_assignments enr
	      JOIN '.$CFG->prefix.'user usr ON usr.id=enr.userid
	      WHERE enr.roleid = '.$role->id.'
	      AND enr.contextid = '.$context->id.'
	      AND enr.enrol = \'lml\' ';
	  if (!empty($ldapmembers)) {
	      $sql .= 'AND usr.idnumber NOT IN (\''. join('\',\'', $ldapmembers).'\')';
	  } else {
	      print ("Empty enrolment for $course_obj->shortname \n");
	  }
	  $todelete = get_records_sql($sql);
	  if(!empty($todelete)){
	    foreach ($todelete as $member) {
	      $member = $member->user;

	      if (role_unassign($role->id, $member, 0, $context->id, 'lml')) {
		  print "Unassigned $type from $member for course $course_obj->id ($course_obj->shortname)\n";
	      } else {
		  print "Failed to unassign $type from $member for course $course_obj->id ($course_obj->shortname)\n";
	      }
	    }
	  }
	    
	  // insert current enrolments 
	  // bad we can't do INSERT IGNORE with postgres...
	  foreach ($ldapmembers as $ldapmember) {
	    $sql = 'SELECT id,1 FROM '.$CFG->prefix.'user '
		    ." WHERE auth='ldap' AND idnumber='$ldapmember'";
	    $member = get_record_sql($sql); 
//            print "sql: $sql \nidnumber = ".stripslashes($ldapmember)." \n".var_dump($member); 
	    if(empty($member) || empty($member->id)){
		print "Could not find user ".stripslashes($ldapmember).", skipping\n";
		continue;
	    }
	    $member = $member->id;
	    if (!get_record('role_assignments', 'roleid', $role->id, 
			  'contextid', $context->id, 
			  'userid', $member, 'enrol', 'lml')){
	      if (role_assign($role->id, $member, 0, $context->id, 0, 0, 0, 'lml')){
		  print "Assigned role $type to $member (".stripslashes($ldapmember).") for course $course_obj->id ($course_obj->shortname)\n";
	      } else {
		  print "Failed to assign role $type to $member (".stripslashes($ldapmember).") for course $course_obj->id ($course_obj->shortname)\n";
	      }
	    }
	  }
	}
      }
    return true;
}

/* sync student enrolments with lml
 * iterate through all moodle courses, get the current enrolments, get the new enrolments from
 * ldap and delete / add the differences
 */
function sync_student_enrolments($ldap_connection, $role, $enrol) {
    global $CFG;

    //error_log('[ENROL_LML sync_student_enrolments] called');
    if(empty($this->teacher_array)) {
		$this->init_teacher_array($ldap_connection);
    }
    
    //only the students role will be used
    if($role->shortname != $CFG->enrol_lml_student_role) {
        error_log('[ENROL_LML sync_student_enrolments] no student, nothing to do!');
	return true;
    }
    
    //iterate over all moodle course ids and use only those with non empty idnumber.
    $courses = get_records_sql('SELECT id,idnumber,shortname FROM '.$CFG->prefix.'course');
    //error_log('[ENROL_LML sync_student_enrolments] courses('.print_r($courses,true).')');
    print "==Syncing students\n";
    foreach($courses as $course){
		if(empty($course->idnumber))
    		continue;
    	$old=$this->get_moodle_course_enrolments($course->id, $role);
    	$new=$this->get_ldap_groups_members_array($ldap_connection, $course->idnumber);
    	$this->create_course_diff_arrays($old, $new, $todelete, $toadd);
    	if(empty($toadd) && empty($todelete))
    		continue;
    	$context = get_context_instance(CONTEXT_COURSE, $course->id);
    	print "   * $course->id ($course->shortname)\n";
    	if(!empty($todelete)){
    		foreach ($todelete as $member) {
    			$mdl_user=get_record('user','idnumber',$member,'auth','ldap');
    			if(empty($mdl_user)){
    				error_log('[ENROL_LML sync_student_enrolments] moodle user '.$member.' not found for enrolment deletion in mdl_user database.');
    			}
    			elseif (role_unassign($role->id, $mdl_user->id, 0, $context->id, 'lml')) {
    				print "Unassigned $role->shortname from $member for course $course->id ($course->shortname)\n";
    			} else {
    				print "Failed to unassign $role->shortname from $member for course $course->id ($course->shortname)\n";
    			}
    		}
    	}
    	if(!empty($toadd)){
    		foreach($toadd as $member){
    			$mdl_user=get_record('user','idnumber',$member,'auth','ldap');
    			if(empty($mdl_user)){
    				error_log('[ENROL_LML sync_student_enrolments] moodle user '.$member.' not found for enrolment in mdl_user database.');
    			}
    			elseif(role_assign($role->id, $mdl_user->id, 0, $context->id, 0, 0, 0, 'lml')){
				    print "Assigned role $role->shortname to $member for course $course->id ($course->shortname)\n";
				} else {
				    print "Failed to assign role $role->shortname to $member for course $course->id ($course->shortname)\n";
				}
    		}
    	}
    }
    return true;
}

/// sync enrolments with lml
function sync_enrolments($type, $enrol = false) {
    global $CFG;
    //error_log('[ENROL_LML sync_enrolments] start type='.$type);
    
    // Get the role. If it doesn't exist, that is bad.
    $role = get_record('role', 'shortname', $type);
    if (!$role) {
        notify("No such role: $type");
        return false;
    }

    // Connect to the external database
    $ldap_connection = $this->enrol_lml_connect();
    if (!$ldap_connection) {
        @ldap_close($ldap_connection);
        notify("LML-module cannot connect to server: $CFG->enrol_lml_host_url");
        return false;
    }
    
    // we are connected OK, continue...
    $this->enrol_lml_bind($ldap_connection);

    // first fix sort order
    fix_course_sortorder();
    
    //sync enrolments in the teachers courses
    if($role->shortname == $CFG->enrol_lml_teachers_role ) {
      $this->sync_teacher_enrolments($ldap_connection, $role, $enrol);
    }
    //sync enrolments in the students courses
    if($role->shortname == $CFG->enrol_lml_student_role ) {
      $this->sync_student_enrolments($ldap_connection, $role, $enrol);
    }
    
    //autoremove teacher course categories of removed teachers if requested
    if($CFG->enrol_lml_teachers_category_autoremove && $role->shortname == $CFG->enrol_lml_teachers_course_role) {
	print "==Syncing categories\n";
	$teacher_obj = $this->get_teacher_category();
	$attic_obj = $this->get_teacher_attic_category();
	if(!empty($teacher_obj) && !empty($attic_obj)){
	    if($categories = get_categories($teacher_obj->id,'name')){
		foreach($categories as $cat){
		    if(!$this->is_teacher($ldap_connection,$cat->name) && !$this->ignored_teacher($cat->name)){ //move category to attic
			if(!move_category($cat, $attic_obj)) {
			    print "* Could not move category for user '$cat->name' to attic.\n";
			}
			else {
			    print "* Removed category of removed teacher '$cat->name'\n";
			}
		    }
		}
	    }
	}
	else {
	    print "* Could not get teacher category($teacher_obj) or attic category($attic_obj)";
	}
    }
    
    // we are done now, a bit of housekeeping
    fix_course_sortorder();
    
    @ldap_close($ldap_connection);
    return true;
}


/// Overide the get_access_icons() function
function get_access_icons($course) {
}


/// Override the base config_form() function
function config_form($frm) {
    global $CFG;

    $this->check_legacy_config();
    
    include("$CFG->dirroot/enrol/lml/config.html");
}

/// Override the base process_config() function
function process_config($config) {

    $this->check_legacy_config();

    if (!isset ($config->enrol_lml_host_url)) {
        $config->enrol_lml_host_url = '';
    }
    set_config('enrol_lml_host_url', $config->enrol_lml_host_url);

    if (!isset ($config->enrol_lml_version)) {
        $config->enrol_lml_version = '';
    }
    set_config('enrol_lml_version', $config->enrol_lml_version);
    
    if (!isset ($config->enrol_lml_bind_dn)) {
        $config->enrol_lml_bind_dn = '';
    }
    set_config('enrol_lml_bind_dn', $config->enrol_lml_bind_dn);

    if (!isset ($config->enrol_lml_bind_pw)) {
        $config->enrol_lml_bind_pw = '';
    }
    set_config('enrol_lml_bind_pw', $config->enrol_lml_bind_pw);    
    
    if (!isset ($config->enrol_lml_search_sub)) {
      $config->enrol_lml_search_sub = '0';
    }
    set_config('enrol_lml_search_sub', $config->enrol_lml_search_sub);

    if (!isset ($config->enrol_lml_global_context)) {
      $config->enrol_lml_global_context = 'dc=linuxmuster,dc=local';
    }
    set_config('enrol_lml_global_context', $config->enrol_lml_global_context);

    // fields for teachers
    //subtree where teacher group can be found
    if(!isset ($config->enrol_lml_teachers_context)) {
      $config->enrol_lml_teachers_context = 'ou=groups';
    }
    set_config('enrol_lml_teachers_context', $config->enrol_lml_teachers_context);

    //object type for teacher group
    if(!isset ($config->enrol_lml_teachers_object)) {
      $config->enrol_lml_teachers_object = 'posixGroup';
    }
    set_config('enrol_lml_teachers_object', $config->enrol_lml_teachers_object);

    //group object attribute for teachers
    if (!isset ($config->enrol_lml_teachers_group_attribute)) {
	$config->enrol_lml_teachers_group_attribute = 'cn';
    }
        set_config('enrol_lml_teachers_group_attribute', $config->enrol_lml_teachers_group_attribute);

    //group object name for teachers
    if (!isset ($config->enrol_lml_teachers_group_name)) {
	$config->enrol_lml_teachers_group_name = 'teachers';
    }
    set_config('enrol_lml_teachers_group_name', $config->enrol_lml_teachers_group_name);

    //attribute in the teacher group
    if(!isset ($config->enrol_lml_teachers_member_attribute)) {
      $config->enrol_lml_teachers_member_attribute = 'memberUid';
    }
    set_config('enrol_lml_teachers_member_attribute', $config->enrol_lml_teachers_member_attribute);

    //moodle course id number to enrol teachers into
    if (!isset ($config->enrol_lml_teachers_course_id)) {
	$config->enrol_lml_teachers_course_id = 'teachers';
    }
    set_config('enrol_lml_teachers_course_id', $config->enrol_lml_teachers_course_id);
    
    //moodle teachers role
    if (!isset ($config->enrol_lml_teachers_role)) {
	$config->enrol_lml_teachers_role = 'student';
    }
    set_config('enrol_lml_teachers_role', $config->enrol_lml_teachers_role);

    //moodle course prefix for courses with teacher members
    if (!isset ($config->enrol_lml_prefix_teacher_members)) {
	$config->enrol_lml_prefix_teacher_members = 'p_teachers_';
    }
    set_config('enrol_lml_prefix_teacher_members', $config->enrol_lml_prefix_teacher_members);

    //fields for teacher courses
    //moodle teachers course region
    if (!isset ($config->enrol_lml_teachers_course_context)) {
	$config->enrol_lml_teachers_course_context = 'Lehrer';
    }
    set_config('enrol_lml_teachers_course_context', $config->enrol_lml_teachers_course_context);

    //moodle teachers course region
    if (!isset ($config->enrol_lml_teachers_course_role)) {
	$config->enrol_lml_teachers_course_role = 'coursecreator';
    }
    set_config('enrol_lml_teachers_course_role', $config->enrol_lml_teachers_course_role);

    //moodle teachers category autocreate 
    if (!isset ($config->enrol_lml_teachers_category_autocreate)) {
	$config->enrol_lml_teachers_category_autocreate = '0';
    }
    set_config('enrol_lml_teachers_category_autocreate', $config->enrol_lml_teachers_category_autocreate);

    //moodle teachers category autoremove 
    if (!isset ($config->enrol_lml_teachers_category_autoremove)) {
	$config->enrol_lml_teachers_category_autoremove = '0';
    }
    set_config('enrol_lml_teachers_category_autoremove', $config->enrol_lml_teachers_category_autoremove);

    //moodle removed teachers category
    if (!isset ($config->enrol_lml_teachers_removed)) {
	$config->enrol_lml_teachers_removed = 'attic';
    }
    set_config('enrol_lml_teachers_removed', $config->enrol_lml_teachers_removed);

    //moodle teachers to ignore
    if (!isset ($config->enrol_lml_teachers_ignore)) {
	$config->enrol_lml_teachers_ignore = 'bib';
    }
    set_config('enrol_lml_teachers_ignore', $config->enrol_lml_teachers_ignore);

    // fields for students
    //subtrees where students groups can be found
    if(!isset ($config->enrol_lml_student_groups_contexts)) {
      $config->enrol_lml_student_groups_contexts = 'ou=groups';
    }
    set_config('enrol_lml_student_groups_contexts', $config->enrol_lml_student_groups_contexts);

    //object type for students groups
    if(!isset ($config->enrol_lml_student_group_object)) {
      $config->enrol_lml_student_group_object = 'posixGroup';
    }
    //students class numbers
    if(!isset ($config->enrol_lml_student_class_numbers)) {
      $config->enrol_lml_student_class_numbers = '5,6,7,8,9,10,11,12,13';
    }
    
    set_config('enrol_lml_student_class_numbers', $config->enrol_lml_student_class_numbers);
    
    //students project prefix
    if(!isset ($config->enrol_lml_student_project_prefix)) {
      $config->enrol_lml_student_project_prefix = 'p_';
    }
    set_config('enrol_lml_student_project_prefix', $config->enrol_lml_student_project_prefix);
    
    //attribute in the students group
    if(!isset ($config->enrol_lml_student_group_attribute)) {
      $config->enrol_lml_student_group_attribute = 'memberUid';
    }
    set_config('enrol_lml_student_group_attribute', $config->enrol_lml_student_group_attribute);

    //role in the moodle course
    if(!isset ($config->enrol_lml_student_role)) {
      $config->enrol_lml_student_role = 'student';
    }
    set_config('enrol_lml_student_role', $config->enrol_lml_student_role);

    //moodle course id number to enrol students into
    if (!isset ($config->enrol_lml_student_course_id)) {
	$config->enrol_lml_student_course_id = 'cn';
    }
    set_config('enrol_lml_student_course_id', $config->enrol_lml_student_course_id);

    //groups to ignore
    if (!isset ($config->enrol_lml_student_ignored_groups)) {
	$config->enrol_lml_student_ignored_groups = 'teachers,attic';
    }
    set_config('enrol_lml_student_ignored_groups', $config->enrol_lml_student_ignored_groups);

    return true;
}

/**
* This function is run by admin/cron.php
* @return void
*/
function cron() {
    global $CFG, $USER, $SITE;

    if (!isset($CFG->last_enrol_lml_cron)) {
        set_config('last_enrol_lml_cron', 0);
    }

    // notify once an hour only - TODO: add some tz handling here, maybe use timestamps
    if ($CFG->last_enrol_lml_cron == date('Ymdh')) {
	return;
    }

    // Start
    print "Starting enrolments for lml enrolments plugin...\n";
    $roles = get_records('role');
    foreach ($roles as $role) {
        $this->sync_enrolments($role->shortname, true);
    }
    
    // Ende
    print "Finished enrolments for lml enrolments plugin.\n";

    set_config('last_enrol_lml_cron', date('Ymdh'));
}

function enrol_lml_connect(){
/// connects to ldap-server
    global $CFG;

    $result = ldap_connect($CFG->enrol_lml_host_url);
    if ($result) {
        if (!empty($CFG->enrol_lml_version)) {
            ldap_set_option($result, LDAP_OPT_PROTOCOL_VERSION, $CFG->enrol_lml_version);
        }

        if (!empty($CFG->enrol_lml_bind_dn)) {
            $bind = ldap_bind( $result,
                                 $CFG->enrol_lml_bind_dn, 
                                 $CFG->enrol_lml_bind_pw );
            if (!$bind) {
                notify("Error in binding to LML server with credentials");
                trigger_error("Error in binding to LML server $!");
            }

        }
	else {
            $bind = ldap_bind( $result );
            if (!$bind) {
                notify("Error in binding to LML server with anonymous bind");
                trigger_error("Error in binding to LML server $!");
            }

	}
        return $result;
    } else {
        notify("LML-module cannot connect to server: $CFG->enrol_lml_host_url");
        return false;
    }
}

function enrol_lml_bind($ldap_connection){
/// makes bind to ldap for searching users
/// uses ldap_bind_dn or anonymous bind
    global $CFG;

    if ( ! empty($CFG->enrol_lml_bind_dn) ){
        //bind with search-user
        if (!ldap_bind($ldap_connection, $CFG->enrol_lml_bind_dn,$CFG->enrol_lml_bind_pw)){
            notify("Error: could not bind lml with lml_bind_dn/pw");
            return false;
        }

    } else {
        //bind anonymously 
        if ( !ldap_bind($ldap_connection)){
            notify("Error: could not bind lml anonymously");
            return false;
        }  
    }

    return true;
}

/* Search LDAP in all contexts for all groups and return flat record with
 * user names.
 */
function get_ldap_groups_members_array($ldap_connection, $groups){
	global $CFG;
	// define the search pattern
	$groups=$this->filter_valid_groups($groups);
	if(empty($groups))
		return Array();
	$ldap_search_pattern = $this->generate_groups_pattern($groups);
	if (!empty($CFG->{'enrol_lml_student_group_object'})){
		$ldap_search_pattern='(&(objectclass='.$CFG->{'enrol_lml_student_group_object'}.')'.$ldap_search_pattern.')';
	}
	
	//get all contexts and look for first matching user
	$lml_contexts = explode(";",$CFG->{'enrol_lml_student_groups_contexts'});
	
	// get all the fields we will want
	$ldap_fields_wanted = array( 'dn',$CFG->{'enrol_lml_student_group_attribute'},$CFG->{'enrol_lml_student_course_id'});
	$member_field="";
	if(array_key_exists('enrol_lml_student_group_attribute', $CFG)
                  && !empty($CFG->{'enrol_lml_student_group_attribute'})){
		$member_field=strtolower($CFG->{'enrol_lml_student_group_attribute'});
	}
	else {
		$member_field='memberuid';
	}
	array_push($ldap_fields_wanted, $member_field);
	if(array_key_exists('enrol_lml_student_course_id', $CFG)
                  && !empty($CFG->{'enrol_lml_student_course_id'})){
		array_push($ldap_fields_wanted,$CFG->{'enrol_lml_student_course_id'});
	}
	foreach ($lml_contexts as $context) {
		$context = trim($context);
		$context = $context.','.$CFG->{'enrol_lml_global_context'};
		//error_log('[ENROL_LML sync_student_enrolments] find groups in context('.$context.')');
		if ($CFG->enrol_lml_search_sub){
			//use ldap_search to find first user from subtree
			$ldap_result = @ldap_search($ldap_connection,
			$context,
			$ldap_search_pattern,
			$ldap_fields_wanted);
	
		} else {
			//search only in this context
			$ldap_result = @ldap_list($ldap_connection,
			$context,
			$ldap_search_pattern,
			$ldap_fields_wanted,0,0);
		}
		// check and push results
		$records = $ldap_result
		? ldap_get_entries($ldap_connection,$ldap_result)
		: array('count' => 0);
		//error_log('[ENROL_LML get_ldap_groups_members_array] records('.print_r($records,true).')');
		// ldap libraries return an odd array, really. fix it:
		$flat_records = array();
		for ($c=0;$c<$records['count'];$c++) {
			//error_log('[ENROL_LML get_ldap_groups_members_array] records('.print_r($records["$c"]["memberuid"],true).')');
			unset($records["$c"]["$member_field"]['count']);
			// remove teachers if no teachers course
			$idnumber = $records["$c"]{$CFG->enrol_lml_student_course_id}[0];
        	$is_teacher_members_course = $this->has_teachers_as_members($idnumber);
            //error_log('[ENROL_LML get_ldap_groups_members_array] idnumber('.$idnumber.') is_teacher_members_course('.($is_teacher_members_course?'true':'false').')');
        	foreach($records["$c"]["$member_field"] as $member){
				if(!$is_teacher_members_course && $this->is_teacher($ldap_connection,$member)){
					//error_log('[ENROL_LML get_ldap_groups_members_array] member('.$member.') ist ein Lehrer und wird nicht eingefügt.');
					continue;
				}
				$flat_records[]=$member;
			}
		}
		//error_log('[ENROL_LML get_ldap_groups_members_array] records('.print_r($flat_records,true).')');
		// free mem -- is there a leak?
		$records=0; $ldap_result=0;
	}
	return $flat_records;
}

function has_teachers_as_members($group){
	global $CFG;
	if(array_key_exists('enrol_lml_prefix_teacher_members', $CFG)
	&& !empty($CFG->{'enrol_lml_prefix_teacher_members'})){
		if(strpos($group,$CFG->{'enrol_lml_prefix_teacher_members'}) === 0)
			return true;
	}
	if(array_key_exists('enrol_lml_teachers_group_name', $CFG)
	&& !empty($CFG->{'enrol_lml_teachers_group_name'})){
		$teachergroup=$CFG->{'enrol_lml_teachers_group_name'};
	}
	else {
		$teachergroup='teachers';
	}
	if($group == $teachergroup)
			return true;
	return false;
}

/* Get the lml enrolments for this role and this courseid as
 * an array of user uids.
 */
function get_moodle_course_enrolments($courseid,$role){
	global $CFG;
	//error_log('[ENROL_LML get_moodle_course_enrolments] start: courseid('.$courseid.') role('.print_r($role,true).')');
	$context = get_context_instance(CONTEXT_COURSE, $courseid);
	$sql = 'SELECT usr.idnumber
			FROM '.$CFG->prefix.'role_assignments enr
			JOIN '.$CFG->prefix.'user usr ON usr.id=enr.userid
			WHERE enr.roleid = '.$role->id.'
			AND enr.contextid = '.$context->id.'
			AND enr.enrol = \'lml\' ';
	$records = get_records_sql($sql);
	//error_log('[ENROL_LML get_moodle_course_enrolments] records('.print_r($records,true).')');
	$flat_records = array();
	foreach($records as $r)
		$flat_records[] = $r->idnumber;
	//error_log('[ENROL_LML get_moodle_course_enrolments] records('.print_r($flat_records,true).')');
	$records = 0;
	return $flat_records;
}

function create_course_diff_arrays($old,$new,&$todelete,&$toadd){
	//error_log('[ENROL_LML create_course_diff_arrays] Start: old('.print_r($old,true).') new('.print_r($new,true).')');
	if(empty($old)){
		$toadd=array_unique($new);
		array_values($new);
		$todelete=array();
		return;
	}
	if(empty($new)){
		$toadd=array();
		$todelete=array_unique($old);
		array_values($old);
		return;
	}
	sort($old);
	$old=array_unique($old);
	array_values($old);
	sort($new);
	$new=array_unique($new);
	array_values($new);
	//error_log('[ENROL_LML create_course_diff_arrays] Sortiert: old('.print_r($old,true).') new('.print_r($new,true).')');	
	$todelete = array_diff($old,$new);
	$toadd = array_diff($new,$old);
	//error_log('[ENROL_LML create_course_diff_arrays] Sortiert: toadd('.print_r($toadd,true).') todelete('.print_r($todelete,true).')');	
}

function moodle_courses_for_group($group){
	$sql = $this->enrol_localcoursefield.' LIKE \'%'.$group.'%\'';
	//error_log('[ENROL_LML syncstudent_enrolments] sql('.$sql.')');
	$courses_objs = false;
        $courses_objs_all = get_records_select( 'course', $sql );
	//error_log('[ENROL_LML syncstudent_enrolments] '.$sql.':'.print_r($courses_objs_all,true));
	$courses_objs = array_filter($courses_objs_all, create_function('$x','
		foreach(explode(",",$x->'.$this->enrol_localcoursefield.') as $n){
			if($n == \''.$group.'\' ){
				return true;
			}
		}
		return false;'));
	$courses_objs_all = 0;
	return $courses_objs;
}

/* reads the comma separated groups list and
 * returns the groups list with all invalid entries
 * removed.
 * pattern '@^(5|6|7|extra|p_.|teachers)'
 * @uses CFG
*/
function filter_valid_groups($groups){
	global $CFG;
	if(!array_key_exists('enrol_lml_student_class_numbers', $CFG)
	|| empty($CFG->{'enrol_lml_student_class_numbers'}))
		return $groups;
	$pattern='@^('.str_replace(",","|",$CFG->{'enrol_lml_student_class_numbers'}).'|p_.|teachers)@';
	//error_log('[ENROL_LML filter_valid_groups] pattern('.$pattern.')');
	$result=explode(",",$groups);
	//error_log('[ENROL_LML filter_valid_groups] groups('.$groups.')');
	$result=array_filter($result, create_function('$x','
		if(preg_match(\''.$pattern.'\',$x)) 
			return true; 
		else { 
			error_log(\'[ENROL_LML filter_valid_groups] invalid group \'.$x.\' is ignored.\');
			return false;
	}'));
	//error_log('[ENROL_LML filter_valid_groups] result('.$result.') ('.implode(",",$result).')');
	return implode(",",$result);
}

function generate_groups_pattern($groups){
	//create the search pattern to search a comma separated list of groups in LDAP
	//error_log('[ENROL_LML] generate_groups_pattern called');
	if(empty($groups)) 
		return "";
	$group_ar = explode(',',$groups);
	$pattern = "";
	foreach($group_ar as $c) {
		$pattern = $pattern.'(cn='.$c.')';
	}
	//error_log('generate_groups_pattern(classes):'.$pattern);
	if(substr_count($pattern,'cn=')>1){
		$pattern='(|'.$pattern.')';
	}
	//error_log('generate_groups_pattern:'.$pattern);
	return $pattern;
}


function generate_class_pattern(){
  //create the search pattern to search all classes and courses in LDAP
  global $CFG;
  
  //error_log('[ENROL_LML] generate_class_pattern called');
  if(!empty($CFG->enrol_lml_student_class_numbers)) {
      $classes = explode(',',$CFG->{'enrol_lml_student_class_numbers'});
      foreach($classes as $c) {
	$pattern = $pattern.'(cn='.$c.'*)';
      }
  }
  //error_log('generated_class_pattern(classes):'.$pattern);
  if(!empty($CFG->enrol_lml_student_project_prefix)) {
      $pattern = $pattern.'(cn='.$CFG->{'enrol_lml_student_project_prefix'}.'*)';
  }
  //error_log('generated_class_pattern(project):'.$pattern);
  if(substr_count($pattern,'cn=')>1){
      $pattern='(|'.$pattern.')';
  }
  //error_log('generated_class_pattern:'.$pattern);
  return $pattern;
}
    
    
function find_ext_student_enrolments ($ldap_connection, $memberuid, $role){
/// role is a record from the mdl_role table
/// return multidimentional array array with of courses (at least dn and idnumber)
/// 

    global $CFG;

    //error_log('[ENROL_LML] find_ext_student_enrolments called');
    if(empty($memberuid)) { // No "idnumber" stored for this user, so no LDAP enrolments
        return array();
    } 

    //default return value
    $courses = array();
    $this->enrol_lml_bind($ldap_connection);

    //get all contexts and look for first matching user
    $ldap_contexts = explode(";",$CFG->{'enrol_lml_student_groups_contexts'});
    
    // get all the fields we will want for the potential course creation
    // as they are light. don't get membership -- potentially a lot of data. 
    $ldap_fields_wanted = array( 'dn',$CFG->{'enrol_lml_student_course_id'});

    // define the search pattern
    $ldap_search_pattern = "(".$CFG->{'enrol_lml_student_group_attribute'}."=".$this->filter_addslashes($memberuid).")";
    if (!empty($CFG->{'enrol_lml_student_group_object'})){ 
        $ldap_search_pattern='(&(objectclass='.$CFG->{'enrol_lml_student_group_object'}.')'.$ldap_search_pattern.$this->generate_class_pattern().')';
    }
    else {
        $ldap_search_pattern='(&'.$ldap_search_pattern.$this->generate_class_pattern().')';
    }

    $courses=array();
    foreach ($ldap_contexts as $context) {
        $context == trim($context);
        if (empty($context)) {
            continue; // next;
        }
        $context = $context.','.$CFG->{'enrol_lml_global_context'};
        if ($CFG->enrol_lml_search_sub){
            //use ldap_search to find first user from subtree
            $ldap_result = ldap_search($ldap_connection, 
                                        $context, 
                                        $ldap_search_pattern,
                                        $ldap_fields_wanted);

        } else {
            //search only in this context
            $ldap_result = ldap_list($ldap_connection, 
                                        $context, 
                                        $ldap_search_pattern,
                                        $ldap_fields_wanted);
        }
 
        // check and push results
        $records = ldap_get_entries($ldap_connection,$ldap_result);        
        // ldap libraries return an odd array, really. fix it:
	$flat_records=array();
        for ($c=0;$c<$records['count'];$c++) {
            $flat_records[] = $records["$c"][strtolower($CFG->{'enrol_lml_student_course_id'})][0]; 
        }
        if (count($flat_records)) {
            $courses = array_merge($courses, $flat_records);
        }
    }

    return $courses;
}

/**
 *  This function finds the user in the teacher array.
 *
 *  @uses $CFG
 */
function is_teacher($ldap_connection,$user) {
	//error_log('[ENROL_LML is_teacher] is called.');
	if(empty($this->teacher_array)){
    	$this->init_teacher_array($ldap_connection);
  	}
  	//find teacher
  	return in_array($user,$this->teacher_array);
}

/**
 *  This function inits the teacher_array once.
 *
 *  @uses $CFG
 */
function init_teacher_array($ldap_connection) {
  global $CFG;

  //error_log("[init_teacher_array] start");  
  
  //Get the LML teacher role
  $this->enrol_lml_bind($ldap_connection);

  //get the context and look for first matching user
  $ldap_contexts = explode(";",$CFG->{'enrol_lml_teachers_context'});

  //get only the memer field
  $ldap_fields_wanted = array( 'dn',$CFG->{'enrol_lml_teachers_member_attribute'});

  // define the search pattern
  $ldap_search_pattern = "(".$CFG->{'enrol_lml_teachers_group_attribute'}."=".$CFG->{'enrol_lml_teachers_group_name'}.")";
  if (!empty($CFG->{'enrol_lml_teachers_object'})){ 
      $ldap_search_pattern='(&(objectclass='.$CFG->{'enrol_lml_teachers_object'}.')'.$ldap_search_pattern.')';
  }

  foreach ($ldap_contexts as $context) {
      $context == trim($context);
      if (empty($context)) {
	  continue; // next;
      }
      $context=$context.','.$CFG->{'enrol_lml_global_context'};

      if ($CFG->enrol_lml_search_sub){
	  //use ldap_search to find first user from subtree
	  $ldap_result = ldap_search($ldap_connection, 
				      $context, 
				      $ldap_search_pattern,
				      $ldap_fields_wanted);

      } else {
	  //search only in this context
	  $ldap_result = ldap_list($ldap_connection, 
				      $context, 
				      $ldap_search_pattern,
				      $ldap_fields_wanted);
      }
      // check and push results
      $records = ldap_get_entries($ldap_connection,$ldap_result);        
      // ldap libraries return an odd array, really. fix it:

      $flat_records=array();
      for ($c=0;$c<$records['count'];$c++) {
	  	$flat_records = $records["$c"][strtolower($CFG->{'enrol_lml_teachers_member_attribute'})]; 
      }
      if (count($flat_records)) {
		$this->teacher_array = array_merge($flat_records,$this->teacher_array);
      }
  }
  //error_log('[ENROL_LML init_teacher_array] teacher_array('.print_r($this->teacher_array,true).')');
  return true;
}

/**
 * This function checks and creates the teacher category.
 *
 * @uses $CFG
 */
function get_teacher_category(){
    global $CFG;
    //create teacher category if needed
    $cat_obj = get_record( 'course_categories', 'name', $CFG->{'enrol_lml_teachers_course_context'}, 'parent', 0);
    if (empty($cat_obj)){ // category doesn't exist
	error_log("[ENROL_LML] Create non-existing teachers course category $CFG->enrol_lml_teachers_course_context \n");
	$newcategory = new stdClass();
	$newcategory->name = $CFG->{'enrol_lml_teachers_course_context'};
	$newcategory->description = 'Bereich für automatische Lehrerkurse.';
	$newcategory->sortorder = 999;
	$newcategory->parent = 0; //top level category
	if (!insert_record('course_categories', $newcategory)) {
	    notify( "Could not insert the new category '$newcategory->name' ");
	}
	$cat_obj = get_record( 'course_categories', 'name', $CFG->{'enrol_lml_teachers_course_context'}, 'parent', 0);
    }
    if(empty($cat_obj)){
      notify( "[ENROL_ML autocreate/autoremove] could not create teacher course context");
    }
    return $cat_obj;
}

/**
 * This function checks and creates the teacher attic category.
 *
 * @uses $CFG
 */
function get_teacher_attic_category(){
    global $CFG;
    $attic_obj = get_record( 'course_categories', 'name', $CFG->{'enrol_lml_teachers_removed'}, 'parent', 0);
    if (empty($attic_obj)){ // category for removed teachers doesn't exist
	error_log("[ENROL_LML] Create non-existing removed teachers category $CFG->enrol_lml_teachers_removed \n");
	$newcategory = new stdClass();
	$newcategory->name = $CFG->{'enrol_lml_teachers_removed'};
	$newcategory->description = 'Kurse gelöschter Lehrer.';
	$newcategory->sortorder = 999;
	$newcategory->parent = 0; //top level category
	if (!insert_record('course_categories', $newcategory)) {
	    notify( "Could not insert the new category '$newcategory->name' ");
	}
	$attic_obj = get_record( 'course_categories', 'name', $CFG->{'enrol_lml_teachers_removed'}, 'parent', 0);
	if(empty($attic_obj)){
	  notify( "[ENROL_ML autocreate/autoremove] could not create removed teachers context");
	}
    }
    return $attic_obj;
}


/**
 * This function resorts the teacher categories alphabetically.
 *
 */
function resort_categories($id){
    if($categories = get_categories($id,'name')){
	begin_sql();
	$count=1;
	foreach($categories as $cat){
	    set_field('course_categories','sortorder',$count,'id',$cat->id);
	    $count++;
	}
	commit_sql();
    }
    return true;
}

/**
 * This function checks if this teacher is to be ignored in autocreate / autoremove.
 *
 * @uses $CFG
 *
 */
function ignored_teacher(&$name) {
    global $CFG;
    $ignored_teachers = explode(',',$CFG->enrol_lml_teachers_ignore);
    if(!empty($ignored_teachers)){
	foreach($ignored_teachers as $teacher) {
	    if($teacher == $name) {
		return true;
	    }
	}
    }
    return false;
}
 
/**
 * This function autocreates / autoremoves teacher course categories if requested.
 *
 * @uses $CFG;
 */
function setup_teacher_category(&$user,$isteacher){
    global $CFG;
    
    //autocreate / autoremove this teachers category
    $ignore_user = $this->ignored_teacher($user->idnumber);
    //error_log("[ENROL_LML autocreate/remove: $user->name autocreate($CFG->enrol_lml_teachers_category_autocreate) isteacher($isteacher) ignore($ignore_user).");
    if($CFG->enrol_lml_teachers_category_autocreate && $isteacher && !$ignore_user) {
	//create category if needed and autocreate
	$teacher_obj = $this->get_teacher_category();
	if(!empty($teacher_obj)) {
	    $cat_obj = get_record('course_categories','name',$user->idnumber, 'parent', $teacher_obj->id);
	    if(empty($cat_obj)) { //create teacher course category
		//error_log("[ENROL_LML autocreate course category for ".print_r($user,true));
		error_log("[ENROL_LML] Creating course category for teacher '$user->idnumber'.");
		$newcategory = new stdClass();
		$newcategory->name = $user->idnumber;
		$newcategory->description = "Kursbereich des Lehrers $user->firstname $user->lastname ($user->idnumber).";
		$newcategory->parent = $teacher_obj->id; //top level category
		$newcategory->depth = $teacher_obj->depth+1;
		if (!$newcategory->id = insert_record('course_categories', $newcategory)) {
		    error_log( "Could not insert the new teacher course category '$newcategory->name' ".print_r($newcategory,true));
		    return false;
		}
		else {
		    $newcategory->path = $teacher_obj->path.'/'.$newcategory->id;
		    if(!update_record('course_categories',$newcategory)) {
			error_log("Could not update the new teacher course category '$newcategory->name'.");
			return false;
		    }
		    $newcategory->context = get_context_instance(CONTEXT_COURSECAT, $newcategory->id);
		    mark_context_dirty($newcategory->context->path);
		    $this->resort_categories($teacher_obj->id);
		}
		//set teachers role to course creator
		$role = get_record('role','shortname',$CFG->{'enrol_lml_teachers_course_role'});
		if(empty($role)){
		    error_log("[ENROL_LML] Could not get teachers course role ($CFG->enrol_lml_teachers_course_role).");
		    return false;
		}
		$context = get_context_instance(CONTEXT_COURSECAT, $newcategory->id);
		if(empty($context)){
		    error_log("[ENROL_LML] Could not get course catetory context for teacher '$user->idnumber'.");
		    return false;
		}
		if(!role_assign($role->id,$user->id,0,$context->id,0,0,0,'lml')){
		    error_log("[ENROL_LML] Could not assign role ($role->id) to user ($user->idnumber) in context ($context->id).");
		    return false;
		}
	    }
	}
	else {
	    error_log("[ENROL_LML] Could not get teacher category for autocreate.");
	    return false;
	}
    }
    else if($CFG->enrol_lml_teachers_category_autoremove && !$isteacher && !$ignore_user ){//remove cateogry if exists and autoremove
	$teacher_obj = $this->get_teacher_category();
	if(!empty($teacher_obj)){
	    $cat_obj = get_record('course_categories','name',$user->idnumber, 'parent', $teacher_obj->id);
	    if(!empty($cat_obj)) { //remove teacher course category
		notify("[ENROL_LML] Move teacher category for user '$user->idnumber' to attic.");
		$attic_obj = $this->get_teacher_attic_category();
		if(!empty($attic_obj)){
		    if(!move_category($cat_obj, $attic_obj)) {
		      notify("[ENROL_ML] Could not move category for user '$user->idnumber' to attic.");
		      return false;
		    }
		}
		else {
		      notify("[ENROL_ML] Could not get attic category.");
		      return false;
		}
	    }
	}
        else {
	    error_log("[ENROL_LML] Could not get teacher category for autoremove.");
	    return false;
	}
    }
    return true;
}
/**
 * This function checks for the presence of old 'legacy' config settings.  If
 * they exist, it corrects them.
 *
 * @uses $CFG
 */
function check_legacy_config () {
    global $CFG;
    //so far nothing to do
}

/**
 * Quote control characters in texts used in ldap filters - see RFC 4515/2254
 *
 * @param string
 */
function filter_addslashes($text) {
    $text = str_replace('\\', '\\5c', $text);
    $text = str_replace(array('*',    '(',    ')',    "\0"),
                        array('\\2a', '\\28', '\\29', '\\00'), $text);
    return $text;
}

} // end of class

?>
